@@ -6,6 +6,11 @@
 
 #define NGX_ESCAPE_REDIS   4
 
+#define REDIS_AUTH_CMD      "*2\r\n$4\r\nauth\r\n"
+#define REDIS_GET_CMD       "*2\r\n$3\r\nget\r\n"
+#define REDIS_SELECT_CMD    "*2\r\n$6\r\nselect\r\n"
+#define REDIS_PLUSOKCRLF    "+OK\r\n"
+#define REDIS_ERR           "$-1"
 
 #include <ngx_config.h>
 #include <ngx_core.h>
@@ -17,6 +22,7 @@ typedef struct {
     ngx_http_upstream_conf_t   upstream;
     ngx_int_t                  index;
     ngx_int_t                  db;
+    ngx_int_t                  auth;
     ngx_uint_t                 gzip_flag;
 } ngx_http_redis_loc_conf_t;
 
@@ -64,14 +70,19 @@ static ngx_command_t  ngx_http_redis_commands[] = {
       0,
       NULL },
 
-#if defined nginx_version && nginx_version >= 8022
     { ngx_string("redis_bind"),
       NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
       ngx_http_upstream_bind_set_slot,
       NGX_HTTP_LOC_CONF_OFFSET,
       offsetof(ngx_http_redis_loc_conf_t, upstream.local),
       NULL },
-#endif
+
+    { ngx_string("redis_socket_keepalive"),
+      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_FLAG,
+      ngx_conf_set_flag_slot,
+      NGX_HTTP_LOC_CONF_OFFSET,
+      offsetof(ngx_http_redis_loc_conf_t, upstream.socket_keepalive),
+      NULL },
 
     { ngx_string("redis_connect_timeout"),
       NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
@@ -108,6 +119,20 @@ static ngx_command_t  ngx_http_redis_commands[] = {
       offsetof(ngx_http_redis_loc_conf_t, upstream.next_upstream),
       &ngx_http_redis_next_upstream_masks },
 
+    { ngx_string("redis_next_upstream_tries"),
+      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
+      ngx_conf_set_num_slot,
+      NGX_HTTP_LOC_CONF_OFFSET,
+      offsetof(ngx_http_redis_loc_conf_t, upstream.next_upstream_tries),
+      NULL },
+
+    { ngx_string("redis_next_upstream_timeout"),
+      NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
+      ngx_conf_set_msec_slot,
+      NGX_HTTP_LOC_CONF_OFFSET,
+      offsetof(ngx_http_redis_loc_conf_t, upstream.next_upstream_timeout),
+      NULL },
+
     { ngx_string("redis_gzip_flag"),
       NGX_HTTP_MAIN_CONF|NGX_HTTP_SRV_CONF|NGX_HTTP_LOC_CONF|NGX_CONF_TAKE1,
       ngx_conf_set_num_slot,
@@ -149,14 +174,11 @@ ngx_module_t  ngx_http_redis_module = {
     NGX_MODULE_V1_PADDING
 };
 
-/* Initialize additional var for hide "Content-Enconding: gzip" header */
-static ngx_str_t  ngx_http_redis_hide_headers[] = {
-    ngx_null_string
-};
-
-static ngx_str_t  ngx_http_redis_key = ngx_string("redis_key");
-static ngx_str_t  ngx_http_redis_db  = ngx_string("redis_db");
+static ngx_str_t  ngx_http_redis_key  = ngx_string("redis_key");
+static ngx_str_t  ngx_http_redis_db   = ngx_string("redis_db");
+static ngx_str_t  ngx_http_redis_auth = ngx_string("redis_auth");
 static ngx_uint_t ngx_http_redis_db_index;
+static ngx_uint_t ngx_http_redis_auth_index;
 
 
 #define NGX_HTTP_REDIS_END   (sizeof(ngx_http_redis_end) - 1)
@@ -185,40 +207,17 @@ ngx_http_redis_handler(ngx_http_request_t *r)
         return NGX_HTTP_INTERNAL_SERVER_ERROR;
     }
 
-#if defined nginx_version && nginx_version >= 8011
     if (ngx_http_upstream_create(r) != NGX_OK) {
-#else
-    rlcf = ngx_http_get_module_loc_conf(r, ngx_http_redis_module);
-
-    u = ngx_pcalloc(r->pool, sizeof(ngx_http_upstream_t));
-    if (u == NULL) {
-#endif
         return NGX_HTTP_INTERNAL_SERVER_ERROR;
     }
 
-#if defined nginx_version && nginx_version >= 8011
     u = r->upstream;
-#endif
 
-#if defined nginx_version && nginx_version >= 8037
     ngx_str_set(&u->schema, "redis://");
-#else
-    u->schema.len = sizeof("redis://") - 1;
-    u->schema.data = (u_char *) "redis://";
-#endif
 
-#if defined nginx_version && nginx_version >= 8011
     u->output.tag = (ngx_buf_tag_t) &ngx_http_redis_module;
-#else
-    u->peer.log = r->connection->log;
-    u->peer.log_error = NGX_ERROR_ERR;
-#endif
 
-#if defined nginx_version && nginx_version >= 8011
     rlcf = ngx_http_get_module_loc_conf(r, ngx_http_redis_module);
-#else
-    u->output.tag = (ngx_buf_tag_t) &ngx_http_redis_module;
-#endif
 
     u->conf = &rlcf->upstream;
 
@@ -228,10 +227,6 @@ ngx_http_redis_handler(ngx_http_request_t *r)
     u->abort_request = ngx_http_redis_abort_request;
     u->finalize_request = ngx_http_redis_finalize_request;
 
-#if defined nginx_version && nginx_version < 8011
-    r->upstream = u;
-#endif
-
     ctx = ngx_palloc(r->pool, sizeof(ngx_http_redis_ctx_t));
     if (ctx == NULL) {
         return NGX_HTTP_INTERNAL_SERVER_ERROR;
@@ -246,9 +241,7 @@ ngx_http_redis_handler(ngx_http_request_t *r)
     u->input_filter = ngx_http_redis_filter;
     u->input_filter_ctx = ctx;
 
-#if defined nginx_version && nginx_version >= 8011
     r->main->count++;
-#endif
 
     ngx_http_upstream_init(r);
 
@@ -259,17 +252,31 @@ ngx_http_redis_handler(ngx_http_request_t *r)
 static ngx_int_t
 ngx_http_redis_create_request(ngx_http_request_t *r)
 {
-    size_t                          len;
+    size_t                          len = 0;
     uintptr_t                       escape;
     ngx_buf_t                      *b;
     ngx_chain_t                    *cl;
     ngx_http_redis_ctx_t           *ctx;
-    ngx_http_variable_value_t      *vv[2];
+    ngx_http_variable_value_t      *vv[3];
     ngx_http_redis_loc_conf_t      *rlcf;
+    u_char                          lenbuf[NGX_INT_T_LEN];
 
     rlcf = ngx_http_get_module_loc_conf(r, ngx_http_redis_module);
 
-    vv[0] = ngx_http_get_indexed_variable(r, ngx_http_redis_db_index);
+    vv[0] = ngx_http_get_indexed_variable(r, ngx_http_redis_auth_index);
+    if (vv[0] == NULL || vv[0]->not_found || vv[0]->len == 0) {
+        ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
+                       "no auth command provided" );
+    } else {
+        len += sizeof(REDIS_AUTH_CMD) + sizeof("$") - 1;
+        len += ngx_sprintf(lenbuf, "%d", vv[0]->len) - lenbuf;
+        len += sizeof(CRLF) - 1 + vv[0]->len;
+        ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
+                       "auth info: %s", vv[0]->data);
+    }
+    len += sizeof(CRLF) - 1;
+
+    vv[1] = ngx_http_get_indexed_variable(r, ngx_http_redis_db_index);
 
     /*
      * If user do not select redis database in nginx.conf by redis_db
@@ -277,30 +284,34 @@ ngx_http_redis_create_request(ngx_http_request_t *r)
      * some overhead in talk with redis, but this way simplify parsing
      * the redis answer in ngx_http_redis_process_header().
      */
-    if (vv[0] == NULL || vv[0]->not_found || vv[0]->len == 0) {
+    if (vv[1] == NULL || vv[1]->not_found || vv[1]->len == 0) {
         ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                        "select 0 redis database" );
-        len = sizeof("select 0") - 1;
+        len += sizeof(REDIS_SELECT_CMD) + sizeof("$1") + sizeof(CRLF) + sizeof("0") - 1;
     } else {
-        len = sizeof("select ") - 1 + vv[0]->len;
+        len += sizeof(REDIS_SELECT_CMD) + sizeof("$") - 1;
+        len += ngx_sprintf(lenbuf, "%d", vv[1]->len) - lenbuf;
+        len += sizeof(CRLF) - 1 + vv[1]->len;
         ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
-                       "select %s redis database", vv[0]->data);
+                       "select %s redis database", vv[1]->data);
     }
     len += sizeof(CRLF) - 1;
 
-    vv[1] = ngx_http_get_indexed_variable(r, rlcf->index);
+    vv[2] = ngx_http_get_indexed_variable(r, rlcf->index);
 
     /* If nginx.conf have no redis_key return error. */
-    if (vv[1] == NULL || vv[1]->not_found || vv[1]->len == 0) {
+    if (vv[2] == NULL || vv[2]->not_found || vv[2]->len == 0) {
         ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                       "the \"$redis_key\" variable is not set");
         return NGX_ERROR;
     }
 
     /* Count have space required escape symbols. */
-    escape = 2 * ngx_escape_uri(NULL, vv[1]->data, vv[1]->len, NGX_ESCAPE_REDIS);
+    escape = 2 * ngx_escape_uri(NULL, vv[2]->data, vv[2]->len, NGX_ESCAPE_REDIS);
 
-    len += sizeof("get ") - 1 + vv[1]->len + escape + sizeof(CRLF) - 1;
+    len += sizeof(REDIS_GET_CMD) + sizeof("$") - 1;
+    len += ngx_sprintf(lenbuf, "%d", vv[2]->len) - lenbuf;
+    len += sizeof(CRLF) - 1 + vv[2]->len + escape + sizeof(CRLF) - 1;
 
     /* Create temporary buffer for request with size len. */
     b = ngx_create_temp_buf(r->pool, len);
@@ -318,9 +329,13 @@ ngx_http_redis_create_request(ngx_http_request_t *r)
 
     r->upstream->request_bufs = cl;
 
-    /* Add "select " for request. */
-    *b->last++ = 's'; *b->last++ = 'e'; *b->last++ = 'l'; *b->last++ = 'e';
-    *b->last++ = 'c'; *b->last++ = 't'; *b->last++ = ' ';
+    /* add "auth " for request */
+    if (vv[0] != NULL && !(vv[0]->not_found) && vv[0]->len != 0) {
+        /* Add "auth " for request. */
+        b->last = ngx_sprintf(b->last, "%s$%d%s", REDIS_AUTH_CMD, vv[0]->len, CRLF);
+        b->last = ngx_copy(b->last, vv[0]->data, vv[0]->len);
+        *b->last++ = CR; *b->last++ = LF;
+    }
 
     /* Get context redis_db from configuration file. */
     ctx = ngx_http_get_module_ctx(r, ngx_http_redis_module);
@@ -331,12 +346,14 @@ ngx_http_redis_create_request(ngx_http_request_t *r)
      * Add "0" as redis number db to request if redis_db undefined,
      * othervise add real number from context.
      */
-    if (vv[0] == NULL || vv[0]->not_found || vv[0]->len == 0) {
+    if (vv[1] == NULL || vv[1]->not_found || vv[1]->len == 0) {
+        b->last = ngx_sprintf(b->last, "%s$1%s", REDIS_SELECT_CMD, CRLF);
         ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                        "select 0 redis database" );
         *b->last++ = '0';
     } else {
-        b->last = ngx_copy(b->last, vv[0]->data, vv[0]->len);
+        b->last = ngx_sprintf(b->last, "%s$%d%s", REDIS_SELECT_CMD, vv[1]->len, CRLF);
+        b->last = ngx_copy(b->last, vv[1]->data, vv[1]->len);
         ctx->key.len = b->last - ctx->key.data;
         ngx_log_debug1(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
                        "select %V redis database", &ctx->key);
@@ -345,10 +362,7 @@ ngx_http_redis_create_request(ngx_http_request_t *r)
     /* Add "\r\n". */
     *b->last++ = CR; *b->last++ = LF;
 
-
-    /* Add "get" command with space. */
-    *b->last++ = 'g'; *b->last++ = 'e'; *b->last++ = 't'; *b->last++ = ' ';
-
+    b->last = ngx_sprintf(b->last, "%s$%d%s", REDIS_GET_CMD, vv[2]->len, CRLF);
     /* Get context redis_key from nginx.conf. */
     ctx = ngx_http_get_module_ctx(r, ngx_http_redis_module);
 
@@ -360,10 +374,10 @@ ngx_http_redis_create_request(ngx_http_request_t *r)
      */
 
     if (escape == 0) {
-        b->last = ngx_copy(b->last, vv[1]->data, vv[1]->len);
+        b->last = ngx_copy(b->last, vv[2]->data, vv[2]->len);
 
     } else {
-        b->last = (u_char *) ngx_escape_uri(b->last, vv[1]->data, vv[1]->len,
+        b->last = (u_char *) ngx_escape_uri(b->last, vv[2]->data, vv[2]->len,
                                             NGX_ESCAPE_REDIS);
     }
 
@@ -377,8 +391,8 @@ ngx_http_redis_create_request(ngx_http_request_t *r)
 
     /*
      * Summary, the request looks like this:
-     * "select $redis_db\r\nget $redis_key\r\n", where
-     * $redis_db and $redis_key are variable's values.
+     * "auth $redis_auth\r\nselect $redis_db\r\nget $redis_key\r\n", where
+     * $redis_auth, $redis_db and $redis_key are variable's values.
      */
 
     return NGX_OK;
@@ -402,6 +416,11 @@ ngx_http_redis_process_header(ngx_http_request_t *r)
     ngx_http_upstream_t       *u;
     ngx_http_redis_ctx_t      *ctx;
     ngx_http_redis_loc_conf_t *rlcf;
+    ngx_http_variable_value_t *vv;
+    ngx_int_t                  no_auth_cmd;
+
+    vv = ngx_http_get_indexed_variable(r, ngx_http_redis_auth_index);
+    no_auth_cmd = (vv == NULL || vv->not_found || vv->len == 0);
 
     c = try = 0;
 
@@ -411,22 +430,27 @@ ngx_http_redis_process_header(ngx_http_request_t *r)
 
     /*
      * Good answer from redis should looks like this:
-     * "+OK\r\n$8\r\n12345678\r\n"
+     * "+OK\r\n+OK\r\n$8\r\n12345678\r\n"
      *
      * Here is:
-     * "+OK" is answer for first command "select 0".
+     * "+OK\r\n+OK\r\n" is answer for first two commands
+     * "auth password" and "select 0"
      * Next two strings are answer for command "get $redis_key", where
      *
      * "$8" is length of following next string and
      * "12345678" is value of $redis_key, the string.
      *
      * So, if the first symbol is:
-     * "+" (good answer) - try to find 2 strings;
+     * "+" (good answer) - try to find 2 or 3 strings;
      * "-" (bad answer) - try to find 1 string;
-     * othervise answer is invalid. 
+     * othervise answer is invalid.
      */
     if (*p == '+') {
-        try = 2;
+        if (no_auth_cmd) {
+            try = 2;
+        } else {
+            try = 3;
+        }
     } else if (*p == '-') {
         try = 1;
     } else {
@@ -474,8 +498,23 @@ found:
     }
 
     /* Compare pointer and good message, if yes move on the pointer */
-    if (ngx_strncmp(p, "+OK\r\n", sizeof("+OK\r\n") - 1) == 0) {
-        p += sizeof("+OK\r\n") - 1;
+    vv = ngx_http_get_indexed_variable(r, ngx_http_redis_auth_index);
+    if (no_auth_cmd) {
+        if (ngx_strncmp(p, REDIS_PLUSOKCRLF, sizeof(REDIS_PLUSOKCRLF) - 1) == 0) {
+            p += sizeof(REDIS_PLUSOKCRLF) - 1;
+        } else {
+            ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
+                           "+OK\\r\\n was expected here");
+        }
+    } else { /* check double of the "+OK\r\n" */
+        if ((ngx_strncmp(p, REDIS_PLUSOKCRLF, sizeof(REDIS_PLUSOKCRLF) - 1) == 0) &&
+            (ngx_strncmp(p + sizeof(REDIS_PLUSOKCRLF) - 1,
+                            REDIS_PLUSOKCRLF, sizeof(REDIS_PLUSOKCRLF) - 1) == 0)) {
+             p += 2 * (sizeof(REDIS_PLUSOKCRLF) - 1);
+        } else {
+            ngx_log_debug0(NGX_LOG_DEBUG_HTTP, r->connection->log, 0,
+                           "+OK\\r\\n+OK\\r\\n was expected here");
+        }
     }
 
     /*
@@ -483,15 +522,13 @@ found:
      * next symbols are length for upcoming key, "-1" means no key.
      * Set 404 and return.
      */
-    if (ngx_strncmp(p, "$-1", sizeof("$-1") - 1) == 0) {
+    if (ngx_strncmp(p, REDIS_ERR, sizeof(REDIS_ERR) - 1) == 0) {
         ngx_log_error(NGX_LOG_INFO, r->connection->log, 0,
                       "key: \"%V\" was not found by redis", &ctx->key);
 
         u->headers_in.status_n = 404;
         u->state->status = 404;
-#if defined nginx_version && nginx_version >= 1001004
         u->keepalive = 1;
-#endif
 
         return NGX_OK;
     }
@@ -508,7 +545,7 @@ found:
         /* if defined gzip_flag... */
         if (rlcf->gzip_flag) {
             /* hash init */
-            h = ngx_list_push(&r->upstream->headers_in.headers);
+            h = ngx_list_push(&r->headers_out.headers);
             if (h == NULL) {
                 return NGX_ERROR;
             }
@@ -517,19 +554,11 @@ found:
              * add Content-Encoding header for future gunzipping
              * with ngx_http_gunzip_filter module
              */
-            h->hash = ngx_hash(ngx_hash(ngx_hash(ngx_hash(
-                                ngx_hash(ngx_hash(ngx_hash(
-                                ngx_hash(ngx_hash(ngx_hash(
-                                ngx_hash(ngx_hash(ngx_hash(
-                                ngx_hash(ngx_hash('c', 'o'), 'n'), 't'), 'e'),
-                                 'n'), 't'), '-'), 'e'), 'n'), 'c'), 'o'),
-                                 'd'), 'i'), 'n'), 'g');
+            h->hash = 1;
+            h->next = NULL;
             ngx_str_set(&h->key, "Content-Encoding");
             ngx_str_set(&h->value, "gzip");
-            h->lowcase_key = (u_char*) "content-encoding";
-#if (NGX_HTTP_GZIP)
-            u->headers_in.content_encoding = h;
-#endif
+            r->headers_out.content_encoding = h;
         }
 
         /* try to find end of string */
@@ -539,13 +568,8 @@ found:
          * get the length of upcoming redis_key value, convert from ascii
          * if the length is empty, return
          */
-#if defined nginx_version && nginx_version < 1001004
-        r->headers_out.content_length_n = ngx_atoof(len, p - len - 1);
-        if (r->headers_out.content_length_n == -1) {
-#else
         u->headers_in.content_length_n = ngx_atoof(len, p - len - 1);
         if (u->headers_in.content_length_n == -1) {
-#endif
             ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                           "redis sent invalid length in response \"%V\" "
                           "for key \"%V\"",
@@ -580,9 +604,6 @@ ngx_http_redis_filter_init(void *data)
 
     u = ctx->request->upstream;
 
-#if defined nginx_version && nginx_version < 1005003
-    u->length += NGX_HTTP_REDIS_END;
-#else
     if (u->headers_in.status_n != 404) {
         u->length = u->headers_in.content_length_n + NGX_HTTP_REDIS_END;
         ctx->rest = NGX_HTTP_REDIS_END;
@@ -590,7 +611,6 @@ ngx_http_redis_filter_init(void *data)
     } else {
         u->length = 0;
     }
-#endif
 
     return NGX_OK;
 }
@@ -609,11 +629,7 @@ ngx_http_redis_filter(void *data, ssize_t bytes)
     u = ctx->request->upstream;
     b = &u->buffer;
 
-#if defined nginx_version && nginx_version < 1001004
-    if (u->length == ctx->rest) {
-#else
     if (u->length == (ssize_t) ctx->rest) {
-#endif
 
         if (ngx_strncmp(b->last,
                    ngx_http_redis_end + NGX_HTTP_REDIS_END - ctx->rest,
@@ -632,11 +648,9 @@ ngx_http_redis_filter(void *data, ssize_t bytes)
         u->length -= bytes;
         ctx->rest -= bytes;
 
-#if defined nginx_version && nginx_version >= 1001004
         if (u->length == 0) {
             u->keepalive = 1;
         }
-#endif
 
         return NGX_OK;
     }
@@ -676,14 +690,12 @@ ngx_http_redis_filter(void *data, ssize_t bytes)
         ngx_log_error(NGX_LOG_ERR, ctx->request->connection->log, 0,
                       "redis sent invalid trailer");
 
-#if defined nginx_version && nginx_version >= 1001004
         b->last = last;
         cl->buf->last = last;
         u->length = 0;
         ctx->rest = 0;
 
         return NGX_OK;
-#endif
     }
 
     ctx->rest -= b->last - last;
@@ -691,11 +703,9 @@ ngx_http_redis_filter(void *data, ssize_t bytes)
     cl->buf->last = last;
     u->length = ctx->rest;
 
-#if defined nginx_version && nginx_version >= 1001004
-        if (u->length == 0) {
-            u->keepalive = 1;
-        }
-#endif
+    if (u->length == 0) {
+       u->keepalive = 1;
+    }
 
     return NGX_OK;
 }
@@ -726,11 +736,7 @@ ngx_http_redis_create_loc_conf(ngx_conf_t *cf)
 
     conf = ngx_pcalloc(cf->pool, sizeof(ngx_http_redis_loc_conf_t));
     if (conf == NULL) {
-#if defined nginx_version && nginx_version >= 8011
         return NULL;
-#else
-        return NGX_CONF_ERROR;
-#endif
     }
 
     /*
@@ -739,13 +745,15 @@ ngx_http_redis_create_loc_conf(ngx_conf_t *cf)
      *     conf->upstream.bufs.num = 0;
      *     conf->upstream.next_upstream = 0;
      *     conf->upstream.temp_path = NULL;
-     *     conf->upstream.uri = { 0, NULL };
-     *     conf->upstream.location = NULL;
      */
 
+    conf->upstream.local = NGX_CONF_UNSET_PTR;
+    conf->upstream.socket_keepalive = NGX_CONF_UNSET;
+    conf->upstream.next_upstream_tries = NGX_CONF_UNSET_UINT;
     conf->upstream.connect_timeout = NGX_CONF_UNSET_MSEC;
     conf->upstream.send_timeout = NGX_CONF_UNSET_MSEC;
     conf->upstream.read_timeout = NGX_CONF_UNSET_MSEC;
+    conf->upstream.next_upstream_timeout = NGX_CONF_UNSET_MSEC;
 
     conf->upstream.buffer_size = NGX_CONF_UNSET_SIZE;
 
@@ -762,13 +770,7 @@ ngx_http_redis_create_loc_conf(ngx_conf_t *cf)
     conf->upstream.intercept_404 = 1;
     conf->upstream.pass_request_headers = 0;
     conf->upstream.pass_request_body = 0;
-
-    /*
-     * initialize additional parameters for hide
-     * "Content-Encoding: gzip" header
-     */
-    conf->upstream.hide_headers = NGX_CONF_UNSET_PTR;
-    conf->upstream.pass_headers = NGX_CONF_UNSET_PTR;
+    conf->upstream.force_ranges = 1;
 
     conf->index = NGX_CONF_UNSET;
     conf->db = NGX_CONF_UNSET;
@@ -781,11 +783,19 @@ ngx_http_redis_create_loc_conf(ngx_conf_t *cf)
 static char *
 ngx_http_redis_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
 {
-    ngx_hash_init_t            hash;
-
     ngx_http_redis_loc_conf_t *prev = parent;
     ngx_http_redis_loc_conf_t *conf = child;
 
+    ngx_conf_merge_ptr_value(conf->upstream.local,
+                              prev->upstream.local, NULL);
+
+    ngx_conf_merge_value(conf->upstream.socket_keepalive,
+                              prev->upstream.socket_keepalive, 0);
+
+    ngx_conf_merge_uint_value(conf->upstream.next_upstream_tries,
+                              prev->upstream.next_upstream_tries,
+0);
+
     ngx_conf_merge_msec_value(conf->upstream.connect_timeout,
                               prev->upstream.connect_timeout, 60000);
 
@@ -795,6 +805,9 @@ ngx_http_redis_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
     ngx_conf_merge_msec_value(conf->upstream.read_timeout,
                               prev->upstream.read_timeout, 60000);
 
+    ngx_conf_merge_msec_value(conf->upstream.next_upstream_timeout,
+                              prev->upstream.next_upstream_timeout, 0);
+
     ngx_conf_merge_size_value(conf->upstream.buffer_size,
                               prev->upstream.buffer_size,
                               (size_t) ngx_pagesize);
@@ -810,18 +823,6 @@ ngx_http_redis_merge_loc_conf(ngx_conf_t *cf, void *parent, void *child)
                                        |NGX_HTTP_UPSTREAM_FT_OFF;
     }
 
-    /* Initialize hash for hide "Content-Encoding" header */
-    hash.max_size = 512;
-    hash.bucket_size = ngx_align(64, ngx_cacheline_size);
-    hash.name = "redis_hide_headers_hash";
-
-    if (ngx_http_upstream_hide_headers_hash(cf, &conf->upstream,
-            &prev->upstream, ngx_http_redis_hide_headers, &hash)
-        != NGX_OK)
-    {
-        return NGX_CONF_ERROR;
-    }
-
     if (conf->upstream.upstream == NULL) {
         conf->upstream.upstream = prev->upstream.upstream;
     }
@@ -880,6 +881,7 @@ ngx_http_redis_pass(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
     }
 
     rlcf->db = ngx_http_get_variable_index(cf, &ngx_http_redis_db);
+    rlcf->auth = ngx_http_get_variable_index(cf, &ngx_http_redis_auth);
 
     return NGX_CONF_OK;
 }
@@ -900,6 +902,7 @@ ngx_http_redis_add_variables(ngx_conf_t *cf)
 {
     ngx_int_t             n;
     ngx_http_variable_t  *var;
+    ngx_http_variable_t  *authvar;
 
     var = ngx_http_add_variable(cf, &ngx_http_redis_db,
                                 NGX_HTTP_VAR_CHANGEABLE);
@@ -916,5 +919,20 @@ ngx_http_redis_add_variables(ngx_conf_t *cf)
 
     ngx_http_redis_db_index = n;
 
+    authvar = ngx_http_add_variable(cf, &ngx_http_redis_auth,
+                                NGX_HTTP_VAR_CHANGEABLE);
+    if (authvar == NULL) {
+        return NGX_ERROR;
+    }
+
+    authvar->get_handler = ngx_http_redis_reset_variable;
+
+    n = ngx_http_get_variable_index(cf, &ngx_http_redis_auth);
+    if (n == NGX_ERROR) {
+        return NGX_ERROR;
+    }
+
+    ngx_http_redis_auth_index = n;
+
     return NGX_OK;
 }
