@@ -130,6 +130,7 @@ ngx_http_vhost_traffic_status_shm_add_node(ngx_http_request_t *r,
         if (node == NULL) {
             shm_info = ngx_pcalloc(r->pool, sizeof(ngx_http_vhost_traffic_status_shm_info_t));
             if (shm_info == NULL) {
+                ngx_shmtx_unlock(&shpool->mutex);
                 return NGX_ERROR;
             }
 
@@ -147,7 +148,8 @@ ngx_http_vhost_traffic_status_shm_add_node(ngx_http_request_t *r,
         vtsn = (ngx_http_vhost_traffic_status_node_t *) &node->color;
 
         node->key = hash;
-        vtsn->len = (u_char) key->len;
+        vtsn->len = key->len;
+        vtsn->ignore_status = vtscf->ignore_status;
         ngx_http_vhost_traffic_status_node_init(r, vtsn);
         vtsn->stat_upstream.type = type;
         ngx_memcpy(vtsn->data, key->data, key->len);
@@ -246,6 +248,19 @@ ngx_http_vhost_traffic_status_shm_add_node_cache(ngx_http_request_t *r,
         return NGX_OK;
     }
 
+    /*
+     * If max_size in proxy_cache_path directive is not specified,
+     * the system dependent value NGX_MAX_OFF_T_VALUE is assigned by default.
+     *
+     * proxy_cache_path ... keys_zone=name:size [max_size=size] ...
+     *
+     *     keys_zone's shared memory size:
+     *         cache->shm_zone->shm.size
+     *
+     *     max_size's size:
+     *         cache->max_size
+     */
+
     if (init == NGX_HTTP_VHOST_TRAFFIC_STATUS_NODE_NONE) {
         vtsn->stat_cache_max_size = (ngx_atomic_uint_t) (cache->max_size * cache->bsize);
 
@@ -467,14 +482,14 @@ ngx_http_vhost_traffic_status_shm_add_upstream(ngx_http_request_t *r)
 
 found:
 
-    state = r->upstream_states->elts;
-    if (state[0].peer == NULL) {
+    state = u->state;
+    if (state->peer == NULL) {
         ngx_log_error(NGX_LOG_ERR, r->connection->log, 0,
                       "shm_add_upstream::peer failed");
         return NGX_ERROR;
     }
 
-    dst.len = (uscf->port ? 0 : uscf->host.len + sizeof("@") - 1) + state[0].peer->len;
+    dst.len = (uscf->port ? 0 : uscf->host.len + sizeof("@") - 1) + state->peer->len;
     dst.data = ngx_pnalloc(r->pool, dst.len);
     if (dst.data == NULL) {
         return NGX_ERROR;
@@ -482,13 +497,13 @@ found:
 
     p = dst.data;
     if (uscf->port) {
-        p = ngx_cpymem(p, state[0].peer->data, state[0].peer->len);
+        p = ngx_cpymem(p, state->peer->data, state->peer->len);
         type = NGX_HTTP_VHOST_TRAFFIC_STATUS_UPSTREAM_UA;
 
     } else {
         p = ngx_cpymem(p, uscf->host.data, uscf->host.len);
         *p++ = NGX_HTTP_VHOST_TRAFFIC_STATUS_KEY_SEPARATOR;
-        p = ngx_cpymem(p, state[0].peer->data, state[0].peer->len);
+        p = ngx_cpymem(p, state->peer->data, state->peer->len);
         type = NGX_HTTP_VHOST_TRAFFIC_STATUS_UPSTREAM_UG;
     }
 
